import os
import calcpis as pis
import evalsolutil as se
import calcdurs as ds


def soleval(csv_file):
    """ Evaluates a solutions and exports the corresponding Performance Indicators."""
    parentdir = input("Input the main directory of the instances:")
    exportdir = input("Input the export directory:")
    sol_rows = se.read_solution(csv_file)  # Reads the solution.csv file
    export_data = list()
    for row in sol_rows:
        path = os.path.join(parentdir, row[0])
        file = row[1] + ".txt"
        path = os.path.join(path, file)
        rooms = eval(row[1].split("_")[2])  # Gathers the number of ORs from the filename of the instance
        sol = se.decoder(row[2])  # Decodes the code for the cost_eval() function
        durations = ds.export_stats(path)  # Reads the file that contains the surgeries durations
        overtime, idletime, makespan = pis.cost_eval(sol, durations, rooms)  # Calculates the PIs
        export_data.append(se.ExpFile(row[0], row[1], overtime, idletime, makespan))  # Keeps the results in a class
    se.csv_exporter(export_data, exportdir)  # Exports in csv
    print("Successfully exported!")




import csv
import os


class ExpFile:  # Class keeping the PIs
    def __init__(self, type, filename, overtime, idletime, makespan):
        self.type = type
        self.filename = filename
        self.overtime = overtime
        self.idletime = idletime
        self.makespan = makespan


def read_solution(csv_file):  # Reads the solution.csv file
    with open(csv_file, "r", newline='') as out:
        data = csv.reader(out, delimiter=",")
        next(data, None)  # Will not input contain the headers
        solution = [[row[0], row[1], eval(row[2])] for row in data]
    return solution


def decoder(sol):  # Decodes to list of lists
    rooms = sol.count("#")
    solution = [[] for i in range(rooms)]
    count = 0
    for i in sol:
        if i == "#":
            count += 1
        else:
            solution[count-1].append(i)
    return solution


def read_instance(instance_path): # Reads the durations from the instance
    print('Reading {} ...'.format(instance_path))
    with open(instance_path, "r") as out:
        data = csv.reader(out)
        next(data, None)  # Will not input contain the headers to the list.
        durations = [eval(row[1]) for row in data] # Reads mean and standard deviation
    return durations


def csv_exporter(export_data, exportdir):  # Exports the results, in predefined path
    path = os.path.join(exportdir, 'pis_output.csv')
    data = open(path, 'w')
    output = csv.writer(data)
    output.writerow(["Type", "Name", "Overtime/OR", "Idletime/OR", 'Makespan'])
    for row in export_data:
        outrow = list()
        outrow.append(row.type)
        outrow.append(row.filename)
        outrow.append(row.overtime)
        outrow.append(row.idletime)
        outrow.append(row.makespan)
        output.writerow(outrow)
    data.close()



def calc_duration(solution, duration, rooms):  # Calculates the ORs final capacity
    total_duration = [0]*rooms
    for room in range(rooms):
        for surgery in range(len(solution[room])):
            if solution[room][surgery] == -1:  # There are ORs that contain 0 surgeries
                total_duration[room] = -480  # -480 means Idletime of 480 minutes
            else:
                surg = solution[room][surgery]
                surg -= 1  # Substracts 1 cause Python starts from 0
                total_duration[room] += duration[surg]
    return total_duration


def calc_overtime(total_duration):  # Calculates the Overtime of the ORs
    over_time = list()
    for room_duration in total_duration:
        if room_duration == -480:
            over_time.append(0)
        else:
            if 480 - room_duration >= 0:
                over_time.append(0)
            else:
                over_time.append(room_duration - 480)
    return sum(over_time)/len(over_time)


def calc_idletime(total_duration):  # Calculates the Idletime of the ORs
    idle_time = list()
    for room_duration in total_duration:
        if room_duration == - 480:
            idle_time.append(480)
        else:
            if 480 - room_duration > 0:
                idle_time.append(480 - room_duration)
            else:
                idle_time.append(0)
    return sum(idle_time)/len(total_duration)


def calc_makespan(total_duration):  # Returns the max value from the Makespan list
    return max(total_duration)


def cost_eval(solution, duration, rooms):  # Main function - returns the PIs
    total_duration = calc_duration(solution, duration, rooms)
    overtime = calc_overtime(total_duration)
    idletime = calc_idletime(total_duration)
    makespan = calc_makespan(total_duration)
    return overtime, idletime, makespan



import math


def get_file_elements(lines):
    characteristics = list()
    parameters = list()
    for line_number, line in enumerate(lines):
        if "OR-days" in line and "Capacity" not in line:
            characteristics.append(float(line[0]))
        elif any(x in line for x in ["P1", "P2",  "P3"]):
            parameters = lines[line_number+1:]
            break
    for elem, param in enumerate(parameters):
        parameters[elem] = param[1:4]
    return parameters, characteristics


def read_file(instance):
    with open(instance, "r") as data_read:
        lines = data_read.readlines()
        return [line.split() for line in lines]


def parse_file(file_name):
    lines = read_file(file_name)
    return get_file_elements(lines)


def calc_mean(m, s, g):
    mean_t = g + math.e**(m+(s**2)/2)
    return mean_t


def export_stats(instance):
    parameters, characteristics = parse_file(instance)
    durations = list()
    for row in parameters:
        durations.append(calc_mean(float(row[0]), float(row[1]), float(row[2])))
    return durations



from setuptools import setup

setup(name='bs_solevaluator',
      version='0.1',
      description='Solution evaluator for the Benchmark Set of the ORSP (by Leeftink and Hans 2016).',
      author='Panagiotis Giannakoulias',
      author_email='panosgia87@hotmail.com',
      license='MIT',
      packages=['bs_solevaluator'],
      zip_safe=False)